package edu.jhu.cs.cisst.algorithms.segmentation.gac;

import no.uib.cipr.matrix.Vector;
import no.uib.cipr.matrix.sparse.AbstractIterativeSolver;
import no.uib.cipr.matrix.sparse.BiCG;
import no.uib.cipr.matrix.sparse.CG;
import no.uib.cipr.matrix.sparse.DefaultIterationMonitor;
import no.uib.cipr.matrix.sparse.FlexCompRowMatrix;
import no.uib.cipr.matrix.sparse.IterationReporter;
import no.uib.cipr.matrix.sparse.IterativeSolverNotConvergedException;
import no.uib.cipr.matrix.sparse.SparseVector;
import edu.jhu.cs.cisst.algorithms.optimize.fmg.SolverResolutionLevels;
import edu.jhu.cs.cisst.algorithms.optimize.fmg.grid.BoundaryGrid;
import edu.jhu.cs.cisst.algorithms.optimize.fmg.grid.ConstBoundaryGrid;
import edu.jhu.cs.cisst.algorithms.optimize.fmg.grid.ConstNoBoundaryGrid;
import edu.jhu.cs.cisst.algorithms.optimize.fmg.grid.NoBoundaryGrid;
import edu.jhu.cs.cisst.algorithms.optimize.fmg.pde.CachingPDE;
import edu.jhu.cs.cisst.algorithms.optimize.fmg.pde.PDE;

public class MumfordShah2D extends CachingPDE implements IterationReporter {
	/** The gradient pyramid. */
	protected float[][][][] gradientPyramid;

	/** The heaviside pyramid. */
	protected float[][][] heavisidePyramid;

	/** The image pyramid. */
	protected float[][][] imagePyramid;

	/** The resolution levels. */
	protected SolverResolutionLevels levels;

	/** The smoothing weight. */
	protected double smoothWeight;

	/** The maximum iterations. */
	protected int maxIterations = 1000;

	public enum Region {
		Foreground, Background
	};

	protected Region region;

	/**
	 * Instantiates a new PDE for vector field diffusion.
	 * 
	 * @param smoothWeight
	 *            the diffusion weight
	 * @param dim
	 *            the dimension
	 * @param gradientPyramid
	 *            the gradient pyramid
	 * @param levels
	 *            the resolution levels
	 * @param maxIterations
	 *            the max iterations
	 */
	public MumfordShah2D(double smoothWeight, Region region,
			float[][][][] gradientPyramid, float[][][] imagePyramid,
			float[][][] heavisidePyramid, SolverResolutionLevels levels,
			int maxIterations) {
		this.gradientPyramid = gradientPyramid;
		this.imagePyramid = imagePyramid;
		this.heavisidePyramid = heavisidePyramid;
		this.region = region;
		this.levels = levels;
		this.maxIterations = maxIterations;
		this.smoothWeight = smoothWeight;
	}

	/**
	 * implements method from PDE.
	 * 
	 * @param level
	 *            the level
	 * 
	 * @return the grid spacing
	 * 
	 * @see PDE#getGridSpacing
	 */
	public double getGridSpacing(int level) {
		return levels.getGridSpacing(level);
	}

	/**
	 * implements method from PDE.
	 * 
	 * @param u
	 *            the u
	 * @param f
	 *            the f
	 * @param x
	 *            the x
	 * @param y
	 *            the y
	 * @param z
	 *            the z
	 * 
	 * @return the double
	 * 
	 * @see PDE#evaluateLHS
	 */
	public double evaluateLHS(ConstBoundaryGrid u, ConstNoBoundaryGrid f,
			int x, int y, int z) {
		if (z == 1) {
			int level = u.getLevel();
			double h = getGridSpacing(level);
			double lhs = 0;
			double ind = 0;
			double dx = gradientPyramid[level][0][x][y];
			double dy = gradientPyramid[level][1][x][y];
			double w = smoothWeight / (h * h);
			if (region == Region.Foreground) {
				double regionWeight = w * (1 - heavisidePyramid[level][x][y]);
				lhs = (u.get(x + 1, y, z)
						* (regionWeight - 0.5 * dx * smoothWeight)
						+ u.get(x - 1, y, z)
						* (regionWeight + 0.5 * dx * smoothWeight)
						+ u.get(x, y + 1, z)
						* (regionWeight - 0.5 * dy * smoothWeight)
						+ u.get(x, y - 1, z)
						* (regionWeight + 0.5 * dy * smoothWeight) - 4
						* regionWeight * u.get(x, y, z) - u.get(x, y, z));
			} else if (region == Region.Background) {
				double regionWeight = w * (heavisidePyramid[level][x][y]);
				lhs = (u.get(x + 1, y, z)
						* (regionWeight + 0.5 * dx * smoothWeight)
						+ u.get(x - 1, y, z)
						* (regionWeight - 0.5 * dx * smoothWeight)
						+ u.get(x, y + 1, z)
						* (regionWeight + 0.5 * dy * smoothWeight)
						+ u.get(x, y - 1, z)
						* (regionWeight - 0.5 * dy * smoothWeight) - 4
						* regionWeight * u.get(x, y, z) - u.get(x, y, z));
			}
			return lhs;

		}
		return 0;

	}

	/**
	 * Gets the index.
	 * 
	 * @param i
	 *            the i
	 * @param j
	 *            the j
	 * @param cols
	 *            the cols
	 * 
	 * @return the index
	 */
	protected static final int getIndex(int i, int j, int cols) {
		return i * cols + j;
	}

	/**
	 * @see edu.jhu.cs.cisst.algorithms.optimize.fmg.pde.PDE#solve(edu.jhu.cs.cisst
	 *      .algorithms.optimize.fmg.grid.ConstBoundaryGrid,
	 *      edu.jhu.cs.cisst.algorithms.optimize.fmg.grid.ConstNoBoundaryGrid)
	 */
	public BoundaryGrid solve(ConstBoundaryGrid u, ConstNoBoundaryGrid f) {
		int rows = u.getRows();
		int cols = u.getCols();
		int N = rows * cols;
		int level = u.getLevel();
		double h = getGridSpacing(level);
		// System.out.println("GRID SPACING "+h+" "+rows+" "+cols);
		FlexCompRowMatrix A = new FlexCompRowMatrix(N, N);
		SparseVector b = new SparseVector(N);
		SparseVector x = new SparseVector(N);
		int index = 0;
		double w = smoothWeight / (h * h);
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
				double dx = gradientPyramid[level][0][i][j];
				double dy = gradientPyramid[level][1][i][j];
				double indicator = heavisidePyramid[level][i][j];
				int v11 = getIndex(i, j, cols);
				int v01 = getIndex(i - 1, j, cols);
				int v21 = getIndex(i + 1, j, cols);
				int v10 = getIndex(i, j - 1, cols);
				int v12 = getIndex(i, j + 1, cols);
				double centerWeight = 0;
				if (region == Region.Foreground) {
					double regionWeight = w * (1 - indicator);

					if (i > 0) {
						A.set(index, v01, regionWeight + 0.5 * dx
								* smoothWeight);
						centerWeight++;
					}
					if (i < rows - 1) {
						A.set(index, v21, regionWeight - 0.5 * dx
								* smoothWeight);
						centerWeight++;
					}
					if (j > 0) {
						A.set(index, v10, regionWeight + 0.5 * dy
								* smoothWeight);
						centerWeight++;
					}
					if (j < cols - 1) {
						A.set(index, v12, regionWeight - 0.5 * dy
								* smoothWeight);
						centerWeight++;
					}
					A.set(index, v11, -centerWeight * regionWeight - 1);
				} else if (region == Region.Background) {
					double regionWeight = w * indicator;
					if (i > 0) {
						A.set(index, v01, regionWeight - 0.5 * dx
								* smoothWeight);
						centerWeight++;
					}
					if (i < rows - 1) {
						A.set(index, v21, regionWeight + 0.5 * dx
								* smoothWeight);
						centerWeight++;
					}
					if (j > 0) {
						A.set(index, v10, regionWeight - 0.5 * dy
								* smoothWeight);
						centerWeight++;
					}
					if (j < cols - 1) {
						A.set(index, v12, regionWeight + 0.5 * dy
								* smoothWeight);
						centerWeight++;
					}
					A.set(index, v11, -centerWeight * regionWeight - 1);
				}
				b.set(index, f.get(i, j, 1));
				// Use current RHS as solution
				x.set(index, 0);
				index++;

			}
		}
		// Use conjugate gradient solver
		AbstractIterativeSolver solver = new BiCG(b);
		solver.getIterationMonitor().setIterationReporter(this);
		((DefaultIterationMonitor) solver.getIterationMonitor())
				.setMaxIterations(maxIterations);
		try {
			solver.solve(A, b, x);
		} catch (IterativeSolverNotConvergedException e) {
			System.err
					.println("Solver did not converge in "
							+ solver.getIterationMonitor().iterations()
							+ " iterations");
		}

		BoundaryGrid solution = (BoundaryGrid) u.newInstance(u.getRows(), u
				.getCols(), u.getSlices(), u.getLevel(), 0);
		index = 0;
		int slices = u.getSlices();
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
					solution.set(i, j, 1, x.get(index));
				index++;
			}
		}
		return solution;

	}

	/**
	 * implements method from PDE.
	 * 
	 * @param u
	 *            the u
	 * @param f
	 *            the f
	 * @param x
	 *            the x
	 * @param y
	 *            the y
	 * @param z
	 *            the z
	 * 
	 * @return the double
	 * 
	 * @see PDE#evaluate
	 */
	public double evaluate(ConstBoundaryGrid u, ConstNoBoundaryGrid f, int x,
			int y, int z) {

		int level = u.getLevel();
		double h = getGridSpacing(level);
		double val = u.get(x, y, z);
		if (z == 1) {
			double dx = gradientPyramid[level][0][x][y];
			double dy = gradientPyramid[level][1][x][y];
			double ind = 0;
			if (region == Region.Foreground) {
				ind = 1 - heavisidePyramid[level][x][y];
				double denom = (4 * smoothWeight * ind + h * h);
				val = (smoothWeight
						* (u.get(x + 1, y, z) * (ind - 0.5 * dx)
								+ u.get(x - 1, y, z) * (ind + 0.5 * dx)
								+ u.get(x, y + 1, z) * (ind - 0.5 * dy) + u
								.get(x, y - 1, z)
								* (ind + 0.5 * dy)) - f.get(x, y, z) * h * h)
						/ denom;
			} else if (region == Region.Background) {
				ind = heavisidePyramid[level][x][y];
				double denom = (4 * smoothWeight * ind + h * h);
				val = (smoothWeight
						* (u.get(x + 1, y, z) * (ind + 0.5 * dx)
								+ u.get(x - 1, y, z) * (ind - 0.5 * dx)
								+ u.get(x, y + 1, z) * (ind + 0.5 * dy) + u
								.get(x, y - 1, z)
								* (ind - 0.5 * dy)) - f.get(x, y, z) * h * h)
						/ denom;
			}
		}

		return val;

	}

	/**
	 * implements method from CachingPDE.
	 * 
	 * @param sx
	 *            the sx
	 * @param sy
	 *            the sy
	 * @param sz
	 *            the sz
	 * @param level
	 *            the level
	 * 
	 * @return the no boundary grid
	 * 
	 * @see CachingPDE#actuallySampleRHS
	 */
	@Override
	protected NoBoundaryGrid actuallySampleRHS(int sx, int sy, int sz, int level) {
		NoBoundaryGrid p = new NoBoundaryGrid(sx, sy, sz, level, 0);
		for (int i = 0; i < sx; i++) {
			for (int j = 0; j < sy; j++) {
				double val = imagePyramid[level][i][j];
				p.set(i, j, 1, -val);
			}
		}
		return p;
	}

	/**
	 * implements method from CachingPDE.
	 * 
	 * @param sx
	 *            the sx
	 * @param sy
	 *            the sy
	 * @param sz
	 *            the sz
	 * @param level
	 *            the level
	 * 
	 * @return the object
	 * 
	 * @see CachingPDE#sampleMatrix
	 */
	@Override
	protected Object actuallySampleMatrix(int sx, int sy, int sz, int level) {
		return (new Double(getGridSpacing(level)));
	}

	/**
	 * @see no.uib.cipr.matrix.sparse.IterationReporter#monitor(double, int)
	 */
	public void monitor(double error, int iter) {

	}

	/**
	 * @see no.uib.cipr.matrix.sparse.IterationReporter#monitor(double,
	 *      no.uib.cipr.matrix.Vector, int)
	 */
	public void monitor(double error, Vector res, int iter) {
		if (iter % 10 == 0)
			System.out.println(iter + ") Residual:" + error);
	}
}
